/* vi:set cindent: */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>

#include <sys/socket.h>
#include <fcntl.h>
#include <netinet/in.h>

#include "pproxy.h"

#ifdef SUNOS
#define NOBLOCK O_NDELAY
#else
#define NOBLOCK O_NONBLOCK
#endif

char *connerr = "HTTP/1.0 400 Error\n\nCannot connect to remote host\n";

int
do_request (int s, struct sockaddr *r_addr, int addr_len)
{
  int on = 1;
  struct request *req;

  if (setsockopt (s, SOL_SOCKET, SO_REUSEADDR, &on, sizeof (on)) < 0)
    {
      perror ("request:setsockopt(SO_REUSEADDR)");
      exit (1);
    }

  req = new_request (s);
  http_get_header (req);
  http_analyze_client_header (req);

  if (http_connect_server (req) < 0)
    {
      free_request (req);
      exit (0);
    }

  http_send_request (req);

  /* Send "POST" data */
  if (req->cmd == CMD_POST)
    http_send_post (req);

  /* Get server response */
  http_get_server_header (req);

  http_analyze_server_header (req);

  http_send_reply (req);

  if (req->cachereq)
    open_cache_file (req);

  /* Write data back to the client */
  http_process_data (req);

  free_request (req);
  
  return 0;
}

int
http_connect_server (struct request *req)
{
  struct sockaddr_in addr;
  unsigned long ipaddr;
  int one = 1;

  printf ("Connecting to %s:%u\n", req->server, req->port);

  ipaddr = resolve (req->server);
  if (ipaddr == INADDR_NONE)
    {
      write (req->client_fd, connerr, strlen (connerr));
      return -1;
    }

  memset (&addr, 0, sizeof (addr));
  addr.sin_addr.s_addr = ipaddr;
  addr.sin_port = htons (req->port);
  addr.sin_family = AF_INET;

  req->server_fd = socket (AF_INET, SOCK_STREAM, 0);
  if (req->server_fd == -1)
    return -1;

  if ((setsockopt (req->server_fd, SOL_SOCKET, SO_REUSEADDR, (void *) &one,
		   sizeof (one))) < 0)
    {
      close (req->server_fd);
      req->server_fd = -1;
      write (req->client_fd, connerr, strlen (connerr));
      return -1;
    }

  /* Connect to the server */
  if (connect (req->server_fd, &addr, sizeof (addr)) < 0)
    {
      printf ("Connect failed\n");
      write (req->client_fd, connerr, strlen (connerr));
      return -1;
    }

  return 0;
}

int
http_send_request (struct request *req)
{
  static char *cmds[4] =
  {"", "HEAD", "GET", "POST"};
  struct http_var *var;

  sockprintf (req->server_fd, "%s %s HTTP/1.0\n", cmds[req->cmd], req->path);
  var = req->client_headers;
  while (var)
    {
      sockprintf (req->server_fd, "%s: %s\n", var->name, var->value);
      var = var->next;
    }
  sockprintf (req->server_fd, "\n");

  return 0;
}

int
http_send_post (struct request *req)
{
  int len;
  char *buf;
  int nread;

  printf ("Sending post data...\n");

  buf = (char *) malloc (10240);
  if (buf == NULL)
    return -1;

  len = atoi (req_get_client_header (req, "Content-length"));

  printf ("len=%u\n", len);

  while ((nread = read (req->client_fd, buf, len > 10240 ? 10240 : len)) > 0)
    {
      len -= nread;
      write (req->server_fd, buf, nread);
      write (0, buf, nread);
    }
  free (buf);
  return 0;
}

int
http_send_reply (struct request *req)
{
  struct http_var *var;

  sockprintf (req->client_fd, "HTTP/1.0 %u Blahblah\n", req->server_code);
  var = req->server_headers;
  while (var)
    {
      sockprintf (req->client_fd, "%s: %s\n", var->name, var->value);
      var = var->next;
    }
  sockprintf (req->client_fd, "\n");

  return 0;
}


int
http_process_data (struct request *req)
{
  char *buf;
  int nread;

  printf ("OK, processing data\n");
  buf = (char *) malloc (10240);
  if (buf == NULL)
    return -1;

  while ((nread = read (req->server_fd, buf, 10240)) > 0)
    {
      printf ("got %u bytes\n", nread);
      write (req->client_fd, buf, nread);
      if (req->cachereq)
	if (req->cache_fd != -1)
	  write (req->cache_fd, buf, nread);
    }

  return 0;
}

struct request *
new_request (int client_fd)
{
  struct request *req;

  req = (struct request *) malloc (sizeof (struct request));
  if (!req)
    {
      return NULL;
    }

  memset (req, 0, sizeof (struct request));

  req->client_fd = client_fd;
  req->server_fd = -1;
  req->cache_fd = -1;
  req->port = 80;
  req->cachereq = 1;

  return req;
}

void
free_request (struct request *req)
{
  if (req == NULL)
    return;

  if (req->client_fd)
    close (req->client_fd);

  if (req->server_fd)
    close (req->server_fd);

  if (req->cache_fd)
    close (req->cache_fd);


  req->client_fd = -1;
  req->server_fd = -1;
  req->cache_fd = -1;
  free (req);
}

int
http_get_header (struct request *req)
{
  char *buffer;
  int len;
  int rc;
  int firstline;

  buffer = (char *) malloc (10240);
  len = 0;
  firstline = 1;

  while (1)
    {
      rc = sockgets (req->client_fd, buffer, 10240);
      if (rc != 0)
	{
	  return -1;
	}

      if (buffer[strlen (buffer) - 1] == '\n')
	buffer[strlen (buffer) - 1] = 0;

      /* process string */
      if (firstline)
	{
	  reqh_firstline (req, buffer);
	  firstline = 0;
	}
      else
	{
	  if (buffer[0] == 0)
	    return -1;
	  reqh_header (req, buffer);
	}
    }				/* while(1) */
}

int
reqh_firstline (struct request *req, char *line)
{
  char *cmd;			/* GET */
  char *url;			/* http://xxx:port/file */
  char *proto;			/* HTTP/1.0 or NULL */
  char *server;			/* www.seznam.cz */
  char *port;			/* :80 */
  char *path;			/* /Veda_a_Technika/abcd/ */

  cmd = line;

  req->cmd = CMD_UNKNOWN;
  req->http_version = HTTP_UNKNOWNVERSION;

  url = cmd;
  while (*url != ' ' && *url != 0)
    url++;
  if (*url == 0)
    {
      return -1;
    }
  *url++ = 0;

  proto = url;
  while (*proto != ' ' && *proto != 0)
    proto++;

  /* Parse protocol identification */
  if (*proto == 0)
    {
      req->http_version = HTTP_09;
    }
  else
    {
      *proto++ = 0;
      //if (!strcmp(proto, "HTTP/1.0") || !strcmp(proto, "HTTP/1.1"))
      req->http_version = HTTP_10;
      //else req->http_version = HTTP_UNKNOWNVERSION;
    }

  /* Parse command */
  if (!strcmp (cmd, "GET"))
    req->cmd = CMD_GET;
  else if (!strcmp (cmd, "POST"))
    req->cmd = CMD_POST;
  else if (!strcmp (cmd, "HEAD"))
    req->cmd = CMD_HEAD;
  else
    req->cmd = CMD_UNKNOWN;

  if (!strncmp (url, "http://", 7))
    {
      server = url + 7;
      port = server;
      while (*port && *port != ':' && *port != '/')
	port++;
      if (*port == 0)
	{
	  req->path = strdup ("/");
	  if (!req->path)
	    return -1;
	  req->server = strdup (server);
	  if (!req->server)
	    return -1;
	  req->port = 80;
	  return 0;
	}
      else if (*port == '/')
	{
	  path = port;
	  req->port = 80;
	  req->path = strdup (path);
	  if (!req->path)
	    return -1;
	  *port = 0;
	  req->server = strdup (server);
	  if (!req->server)
	    return -1;
	  return 0;
	}
      else
	{
	  *port++ = 0;
	  req->server = strdup (server);
	  if (req->server == NULL)
	    return -1;
	  path = port;
	  while (*path && *path != '/')
	    path++;
	  if (*path == 0)
	    {
	      req->path = strdup ("/");
	      if (!req->path)
		return -1;
	      req->port = atoi (port);
	      return 0;
	    }
	  else
	    {
	      *path++ = 0;
	      req->path = strdup (path);
	      if (!req->path)
		return -1;
	      req->port = atoi (port);
	      return 0;
	    }
	}
    }
  else
    {
      req->path = strdup (url);
      if (!req->path)
	return -1;
    }

  return 0;
}

int
reqh_header (struct request *req, char *line)
{
  char *name;
  char *value;

  name = line;
  value = name;
  while (*value && *value != ':')
    value++;
  if (*value == 0)
    return -1;
  *value++ = 0;
  while (*value == ' ')
    value++;
  printf ("CLI:%s=%s\n", name, value);
  return req_set_client_header (req, name, value);
}

int
http_get_server_header (struct request *req)
{
  char *buffer;
  int len;
  int rc;
  int firstline;

  buffer = (char *) malloc (10240);
  len = 0;
  firstline = 1;

  while (1)
    {
      rc = sockgets (req->server_fd, buffer, 10240);
      if (rc != 0)
	{
	  return -1;
	}

      if (buffer[strlen (buffer) - 1] == '\n')
	buffer[strlen (buffer) - 1] = 0;

      /* process string */
      if (firstline)
	{
	  reqh_response (req, buffer);
	  firstline = 0;
	}
      else
	{
	  if (buffer[0] == 0)
	    return -1;
	  reqh_server_header (req, buffer);
	}
    }				/* while(1) */

}

int
reqh_response (struct request *req, char *line)
{
  char *proto;			/* HTTP/1.0 */
  char *error;			/* 404 */
  char *message;		/* Not found */

  proto = line;

  error = proto;
  while (*error != ' ' && *proto != 0)
    error++;
  if (*error == 0)
    {
      return -1;
    }
  *error++ = 0;

  message = error;
  while (*message != ' ' && *message != 0)
    message++;

  if (*message != 0)
    *message++ = 0;

  printf ("%s: %s\n", error, message);

  req->server_code = atoi (error);

  return 0;
}

int
reqh_server_header (struct request *req, char *line)
{
  char *name;
  char *value;

  name = line;
  value = name;
  while (*value && *value != ':')
    value++;
  if (*value == 0)
    return -1;
  *value++ = 0;
  while (*value == ' ')
    value++;
  printf ("SRV: %s=%s\n", name, value);
  return req_set_server_header (req, name, value);
}

int
http_analyze_client_header (struct request *req)
{
  /* TODO: If req->server not specified, get it from Host header */
  req_set_client_header (req, "Connection", "close");
  req_set_client_header (req, "Host", req->server);

  if (req->cmd != CMD_GET)
    req->cachereq = 0;
  return 0;
}

int
http_analyze_server_header (struct request *req)
{
  req_set_server_header (req, "Connection", "close");
  req_set_server_header (req, "Proxy-Connection", "close");
  if (req->server_code != 200)
    req->cachereq = 0;
  return 0;
}

int
open_cache_file (struct request *req)
{
  char *path;
  char *dir;
  char port[10];

  path = (char *) malloc (20 + strlen (req->server) + 1 + 5 + strlen (req->path)
			  + 20);
  if (!path)
    return -1;

  strcpy (path, "/cache/");
  strcat (path, req->server);
  if (req->port != 80)
    {
      sprintf (port, ":%u", req->port);
      strcat (path, port);
    }
  if (req->path[0] != '/')
    strcat (path, "/");
  strcat (path, req->path);
  if (req->path[strlen (req->path) - 1] == '/')
    strcat (path, "index.html");

  printf ("Storing in %s\n", path);

  req->cache_file = path;

  dir = strdup (path);
  if (!dir)
    return -1;

  while (dir[0] != 0 && dir[strlen (dir) - 1] != '/')
    dir[strlen (dir) - 1] = 0;

  if (dir[strlen (dir) - 1] == '/')
    dir[strlen (dir) - 1] = 0;

  maketree (dir);
  free (dir);

  req->cache_fd = open (req->cache_file, O_CREAT | O_TRUNC | O_RDWR, 0644);
  if (!req->cache_fd)
    {
      perror (req->cache_file);
      return -1;
    }
  return 0;

}
